import autosubmitAPIwu.experiment.common_requests as ExperimentUtils


def main():
    """
    Updates STATUS of experiments.
    """
    ExperimentUtils.update_running_experiments()


if __name__ == "__main__":
    main()



#!/usr/bin/env python

# Copyright 2014 Climate Forecasting Unit, IC3

# This file is part of Autosubmit.

# Autosubmit is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Autosubmit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Autosubmit.  If not, see <http: www.gnu.org / licenses / >.


import time

import os

from bscearth.utils.log import Log

from autosubmitAPIwu.config.basicConfig import BasicConfig
from autosubmitAPIwu.config.config_common import AutosubmitConfig
from submitter import Submitter
from autosubmitAPIwu.platforms.psplatform import PsPlatform
from autosubmitAPIwu.platforms.lsfplatform import LsfPlatform
from autosubmitAPIwu.platforms.pbsplatform import PBSPlatform
from autosubmitAPIwu.platforms.sgeplatform import SgePlatform
from autosubmitAPIwu.platforms.ecplatform import EcPlatform
from autosubmitAPIwu.platforms.slurmplatform import SlurmPlatform
from autosubmitAPIwu.platforms.locplatform import LocalPlatform
from autosubmitAPIwu.platforms.paramiko_platform import ParamikoPlatformException


class ParamikoSubmitter(Submitter):
    """
    Class to manage the experiments platform
    """

    def load_platforms(self, asconf, retries=5):
        """
        Create all the platforms object that will be used by the experiment

        :param retries: retries in case creation of service fails
        :param asconf: autosubmit config to use
        :type asconf: AutosubmitConfig
        :return: platforms used by the experiment
        :rtype: dict
        """

        platforms_used = list()
        hpcarch = asconf.get_platform()

        job_parser = asconf.jobs_parser
        for job in job_parser.sections():
            hpc = job_parser.get_option(job, 'PLATFORM', hpcarch).lower()
            if hpc not in platforms_used:
                platforms_used.append(hpc)

        parser = asconf.platforms_parser

        platforms = dict()
        local_platform = LocalPlatform(asconf.expid, 'local', BasicConfig)
        local_platform.max_wallclock = asconf.get_max_wallclock()
        local_platform.max_processors = asconf.get_max_processors()
        local_platform.max_waiting_jobs = asconf.get_max_waiting_jobs()
        local_platform.total_jobs = asconf.get_total_jobs()
        local_platform.scratch = os.path.join(BasicConfig.LOCAL_ROOT_DIR, asconf.expid, BasicConfig.LOCAL_TMP_DIR)
        local_platform.temp_dir = os.path.join(BasicConfig.LOCAL_ROOT_DIR, 'ASlogs')
        local_platform.root_dir = os.path.join(BasicConfig.LOCAL_ROOT_DIR, local_platform.expid)
        local_platform.host = 'localhost'
        platforms['local'] = local_platform
        platforms['LOCAL'] = local_platform

        for section in parser.sections():

            if section.lower() not in platforms_used:
                continue

            platform_type = parser.get_option(section, 'TYPE', '').lower()
            platform_version = parser.get_option(section, 'VERSION', '')
            try:
                if platform_type == 'pbs':
                    remote_platform = PBSPlatform(asconf.expid, section.lower(), BasicConfig, platform_version)
                elif platform_type == 'sge':
                    remote_platform = SgePlatform(asconf.expid, section.lower(), BasicConfig)
                elif platform_type == 'ps':
                    remote_platform = PsPlatform(asconf.expid, section.lower(), BasicConfig)
                elif platform_type == 'lsf':
                    remote_platform = LsfPlatform(asconf.expid, section.lower(), BasicConfig)
                elif platform_type == 'ecaccess':
                    remote_platform = EcPlatform(asconf.expid, section.lower(), BasicConfig, platform_version)
                elif platform_type == 'slurm':
                    remote_platform = SlurmPlatform(asconf.expid, section.lower(), BasicConfig)
                else:
                    raise Exception("Queue type not specified on platform {0}".format(section))

            except ParamikoPlatformException as e:
                Log.error("Queue exception: {0}".format(e.message))
                return None

            remote_platform.type = platform_type
            remote_platform._version = platform_version

            if parser.get_option(section, 'ADD_PROJECT_TO_HOST', '').lower() == 'true':
                host = '{0}-{1}'.format(parser.get_option(section, 'HOST', None),
                                        parser.get_option(section, 'PROJECT', None))
            else:
                host = parser.get_option(section, 'HOST', None)

            remote_platform.host = host
            remote_platform.max_wallclock = parser.get_option(section, 'MAX_WALLCLOCK',
                                                              asconf.get_max_wallclock())
            remote_platform.max_processors = parser.get_option(section, 'MAX_PROCESSORS',
                                                               asconf.get_max_processors())
            remote_platform.max_waiting_jobs = int(parser.get_option(section, 'MAX_WAITING_JOBS',
                                                                     asconf.get_max_waiting_jobs()))
            remote_platform.total_jobs = int(parser.get_option(section, 'TOTAL_JOBS',
                                                               asconf.get_total_jobs()))
            remote_platform.hyperthreading = parser.get_option(section, 'HYPERTHREADING',
                                                               'false').lower()
            remote_platform.project = parser.get_option(section, 'PROJECT', None)
            remote_platform.budget = parser.get_option(section, 'BUDGET', remote_platform.project)
            remote_platform.reservation = parser.get_option(section, 'RESERVATION', '')
            remote_platform.exclusivity = parser.get_option(section, 'EXCLUSIVITY', '').lower()
            remote_platform.user = parser.get_option(section, 'USER', None)
            remote_platform.scratch = parser.get_option(section, 'SCRATCH_DIR', None)
            remote_platform.temp_dir = parser.get_option(section, 'TEMP_DIR', None)
            remote_platform._default_queue = parser.get_option(section, 'QUEUE', None)
            remote_platform._serial_queue = parser.get_option(section, 'SERIAL_QUEUE', None)
            remote_platform.processors_per_node = parser.get_option(section, 'PROCESSORS_PER_NODE',
                                                                    None)
            remote_platform.custom_directives = parser.get_option(section, 'CUSTOM_DIRECTIVES',
                                                                  None)
            Log.debug("Custom directives from platform.conf: {0}".format(remote_platform.custom_directives))
            remote_platform.scratch_free_space = parser.get_option(section, 'SCRATCH_FREE_SPACE',
                                                                   None)
            remote_platform.root_dir = os.path.join(remote_platform.scratch, remote_platform.project,
                                                    remote_platform.user, remote_platform.expid)
            remote_platform.update_cmds()
            platforms[section.lower()] = remote_platform

        for section in parser.sections():
            if parser.has_option(section, 'SERIAL_PLATFORM'):
                platforms[section.lower()].serial_platform = platforms[parser.get_option(section,
                                                                                         'SERIAL_PLATFORM',
                                                                                         None).lower()]

        self.platforms = platforms



#!/usr/bin/env python

# Copyright 2015 Earth Sciences Department, BSC-CNS

# This file is part of Autosubmit.

# Autosubmit is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Autosubmit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Autosubmit.  If not, see <http://www.gnu.org/licenses/>.
import datetime
import os

import time
from os import path
from os import chdir
from os import listdir
from os import remove

import pydot as pydotplus
import copy

import subprocess

from autosubmitAPIwu.job.job_common import Status
from autosubmitAPIwu.config.basicConfig import BasicConfig
from autosubmitAPIwu.config.config_common import AutosubmitConfig
from bscearth.utils.log import Log
from bscearth.utils.config_parser import ConfigParserFactory

# from diagram import create_bar_diagram


class Monitor:
    """Class to handle monitoring of Jobs at HPC."""
    _table = dict([(Status.UNKNOWN, 'white'), (Status.WAITING, 'gray'), (Status.READY, 'lightblue'), (Status.PREPARED, 'lightsalmon'),
                   (Status.SUBMITTED, 'cyan'), (Status.HELD,
                                                'salmon'), (Status.QUEUING, 'lightpink'), (Status.RUNNING, 'green'),
                   (Status.COMPLETED, 'yellow'), (Status.FAILED, 'red'), (Status.SUSPENDED, 'orange'), (Status.SKIPPED, 'lightyellow')])

    @staticmethod
    def color_status(status):
        """
        Return color associated to given status

        :param status: status
        :type status: Status
        :return: color
        :rtype: str
        """
        if status == Status.WAITING:
            return Monitor._table[Status.WAITING]
        elif status == Status.READY:
            return Monitor._table[Status.READY]
        elif status == Status.PREPARED:
            return Monitor._table[Status.PREPARED]
        elif status == Status.SUBMITTED:
            return Monitor._table[Status.SUBMITTED]
        elif status == Status.QUEUING:
            return Monitor._table[Status.QUEUING]
        elif status == Status.RUNNING:
            return Monitor._table[Status.RUNNING]
        elif status == Status.COMPLETED:
            return Monitor._table[Status.COMPLETED]
        elif status == Status.HELD:
            return Monitor._table[Status.HELD]
        elif status == Status.FAILED:
            return Monitor._table[Status.FAILED]
        elif status == Status.SUSPENDED:
            return Monitor._table[Status.SUSPENDED]
        elif status == Status.SKIPPED:
            return Monitor._table[Status.SKIPPED]
        else:
            return Monitor._table[Status.UNKNOWN]

    def simplified_tree(self, expid, list_of_jobs):
        """
        Create a graph from list of jobs.
        Simplified version of create_tree_list
        """
        graph = pydotplus.Dot(graph_type="digraph")
        already_in = list()
        # Add all nodes
        for job in list_of_jobs:
            if job.name not in already_in:
                already_in.append(job.name)
                node_job = pydotplus.Node(job.name, shape='box', style="filled",
                                          fillcolor=self.color_status(job.status))
                graph.add_node(node_job)
            else:
                print("Repetition Node")

        already_in_edge = list()
        # Add all edges
        for job in list_of_jobs:
            children = job._children
            for child in job._children:
                if (job.name, child.name) not in already_in_edge:
                    already_in_edge.append((job.name, child.name))
                    graph.add_edge(pydotplus.Edge(job.name, child.name))
                else:
                    print("Repetition Edge")
        return graph

    def create_tree_list(self, expid, joblist, packages, groups, hide_groups=False):
        """
        Create graph from joblist

        :param expid: experiment's identifier
        :type expid: str
        :param joblist: joblist to plot
        :type joblist: JobList
        :return: created graph
        :rtype: pydotplus.Dot
        """
        Log.debug('Creating workflow graph...')
        graph = pydotplus.Dot(graph_type='digraph')

        # Log.debug('Creating legend...')
        # legend = pydotplus.Subgraph(
        #     graph_name='Legend', label='Legend', rank="source")
        # legend.add_node(pydotplus.Node(name='UNKNOWN', shape='box', style="",
        #                                fillcolor=self._table[Status.UNKNOWN]))
        # legend.add_node(pydotplus.Node(name='WAITING', shape='box', style="filled",
        #                                fillcolor=self._table[Status.WAITING]))
        # legend.add_node(pydotplus.Node(name='READY', shape='box', style="filled",
        #                                fillcolor=self._table[Status.READY]))
        # legend.add_node(pydotplus.Node(name='SUBMITTED', shape='box', style="filled",
        #                                fillcolor=self._table[Status.SUBMITTED]))
        # legend.add_node(pydotplus.Node(name='QUEUING', shape='box', style="filled",
        #                                fillcolor=self._table[Status.QUEUING]))
        # legend.add_node(pydotplus.Node(name='RUNNING', shape='box', style="filled",
        #                                fillcolor=self._table[Status.RUNNING]))
        # legend.add_node(pydotplus.Node(name='COMPLETED', shape='box', style="filled",
        #                                fillcolor=self._table[Status.COMPLETED]))
        # legend.add_node(pydotplus.Node(name='FAILED', shape='box', style="filled",
        #                                fillcolor=self._table[Status.FAILED]))
        # legend.add_node(pydotplus.Node(name='SUSPENDED', shape='box', style="filled",
        #                                fillcolor=self._table[Status.SUSPENDED]))
        # graph.add_subgraph(legend)

        exp = pydotplus.Subgraph(
            graph_name='Experiment', label=expid, name="maingroup")
        self.nodes_ploted = set()
        Log.debug('Creating job graph...')

        jobs_packages_dict = dict()
        if packages != None and packages:
            for (exp_id, package_name, job_name) in packages:
                jobs_packages_dict[job_name] = package_name

        packages_subgraphs_dict = dict()
        date_member_cluster = dict()
        # print("Iteration joblist: ")
        for job in joblist:
            if job.date != None and job.member != None and job.has_parents():
                date_member_cluster[job.name] = str(
                    str(job.long_name[0:13]) + str(job.member))
                # date_member_cluster[(job.long_name[0:13], job.member)].append(job.name)
            # print(job)
            # print(job.has_parents())
            if job.has_parents():
                continue

            if not groups or job.name not in groups['jobs'] or (job.name in groups['jobs'] and len(groups['jobs'][job.name]) == 1):
                node_job = pydotplus.Node(job.name, shape='box', style="filled",
                                          fillcolor=self.color_status(job.status))

                if groups and job.name in groups['jobs']:
                    group = groups['jobs'][job.name][0]
                    node_job.obj_dict['name'] = group
                    node_job.obj_dict['attributes']['fillcolor'] = self.color_status(
                        groups['status'][group])
                    node_job.obj_dict['attributes']['status'] = groups['status'][group]
                    node_job.obj_dict['attributes']['shape'] = 'box3d'

                exp.add_node(node_job)
                self._add_children(job, exp, node_job, groups, hide_groups)

        if groups:
            #print("In groups.")
            if not hide_groups:
                # print("Not hide groups.")
                for job, group in groups['jobs'].items():
                    #print("Length of group: " + str(len(group)))
                    if len(group) > 1:
                        group_name = 'cluster_' + '_'.join(group)
                        #print("Group name: " + group_name)
                        if group_name not in graph.obj_dict['subgraphs']:
                            subgraph = pydotplus.Cluster(
                                graph_name='_'.join(group))
                            subgraph.obj_dict['attributes']['color'] = 'invis'
                        else:
                            subgraph = graph.get_subgraph(group_name)[0]

                        previous_node = exp.get_node(group[0])[0]
                        #print("Previous Node: " + previous_node)
                        if len(subgraph.get_node(group[0])) == 0:
                            subgraph.add_node(previous_node)

                        for i in range(1, len(group)):
                            node = exp.get_node(group[i])[0]
                            if len(subgraph.get_node(group[i])) == 0:
                                subgraph.add_node(node)

                            edge = subgraph.get_edge(
                                node.obj_dict['name'], previous_node.obj_dict['name'])
                            if len(edge) == 0:
                                edge = pydotplus.Edge(previous_node, node)
                                edge.obj_dict['attributes']['dir'] = 'none'
                                # constraint false allows the horizontal alignment
                                edge.obj_dict['attributes']['constraint'] = 'false'
                                edge.obj_dict['attributes']['penwidth'] = 4
                                subgraph.add_edge(edge)

                            previous_node = node
                        if group_name not in graph.obj_dict['subgraphs']:
                            graph.add_subgraph(subgraph)
            else:
                for edge in copy.deepcopy(exp.obj_dict['edges']):
                    if edge[0].replace('"', '') in groups['status']:
                        del exp.obj_dict['edges'][edge]

            graph.set_strict(True)

        graph.add_subgraph(exp)
        # nodes_in_wrapper = list()
        # not_to_fake = list()
        # for node in exp.get_nodes():
        #     name = node.obj_dict['name']
        #     if name in jobs_packages_dict:
        #         package = jobs_packages_dict[name]
        #         # print(package)
        #         if package not in packages_subgraphs_dict:
        #             packages_subgraphs_dict[package] = pydotplus.Cluster(
        #                 graph_name=package)
        #             packages_subgraphs_dict[package].obj_dict['attributes']['color'] = 'black'
        #             packages_subgraphs_dict[package].obj_dict['attributes']['style'] = 'dashed'
        #         packages_subgraphs_dict[package].add_node(node)
        #         # nodes_in_wrapper.append(name)
        # print(len(date_member_cluster.keys()))
        # for node in exp.get_nodes():
        #     name = node.obj_dict['name']
        #     if name in date_member_cluster.keys():
        #         name_fake_group = date_member_cluster[name]
        #         if name_fake_group not in packages_subgraphs_dict:
        #             packages_subgraphs_dict[name_fake_group] = pydotplus.graphviz.Cluster(graph_name=name_fake_group)
        #             packages_subgraphs_dict[name_fake_group].obj_dict['attributes']['color'] = 'black'
        #             packages_subgraphs_dict[name_fake_group].obj_dict['attributes']['style'] = 'dashed'
        #         packages_subgraphs_dict[name_fake_group].add_node(node)
        #         if name in nodes_in_wrapper:
        #             not_to_fake.append(name_fake_group)
        # print("Done")

        # for package, cluster in packages_subgraphs_dict.items():
        #     if package not in not_to_fake:
        #         # print(package)
        #         # print(cluster)
        #         graph.add_subgraph(cluster)

        # for name in date_member_cluster.keys():
        #     for h in date_member_cluster[name]:
        #         print(str(name) + " : " + str(h))

        # for u in exp.get_node_list():
        #     print(u.obj_dict['name'])

        # for v in exp.get_edge_list():
        #     print(str(v.get_source()) + " to " + str(v.get_destination()))
        #print("Definition finalized")
        Log.debug('Graph definition finalized')
        return graph

    def _add_children(self, job, exp, node_job, groups, hide_groups):
        if job in self.nodes_ploted:
            return
        self.nodes_ploted.add(job)
        if job.has_children() != 0:
            for child in sorted(job.children, key=lambda k: k.name):
                node_child, skip = self._check_node_exists(
                    exp, child, groups, hide_groups)
                if len(node_child) == 0 and not skip:
                    node_child = self._create_node(child, groups, hide_groups)
                    if node_child:
                        exp.add_node(node_child)
                        exp.add_edge(pydotplus.Edge(node_job, node_child))
                    else:
                        skip = True
                elif not skip:
                    node_child = node_child[0]
                    exp.add_edge(pydotplus.Edge(node_job, node_child))
                    skip = True
                if not skip:
                    self._add_children(
                        child, exp, node_child, groups, hide_groups)

    def _check_node_exists(self, exp, job, groups, hide_groups):
        skip = False
        if groups and job.name in groups['jobs']:
            group = groups['jobs'][job.name][0]
            node = exp.get_node(group)
            if len(groups['jobs'][job.name]) > 1 or hide_groups:
                skip = True
        else:
            node = exp.get_node(job.name)

        return node, skip

    def _create_node(self, job, groups, hide_groups):
        node = None

        if groups and job.name in groups['jobs'] and len(groups['jobs'][job.name]) == 1:
            if not hide_groups:
                group = groups['jobs'][job.name][0]
                node = pydotplus.Node(group, shape='box3d', style="filled",
                                      fillcolor=self.color_status(groups['status'][group]))
                node.set_name(group.replace('"', ''))

        elif not groups or job.name not in groups['jobs']:
            node = pydotplus.Node(job.name, shape='box', style="filled",
                                  fillcolor=self.color_status(job.status))
        return node

    def generate_output(self, expid, joblist, path, output_format="pdf", packages=None, show=False, groups=dict(), hide_groups=False):
        """
        Plots graph for joblist and stores it in a file

        :param expid: experiment's identifier
        :type expid: str
        :param joblist: joblist to plot
        :type joblist: JobList
        :param output_format: file format for plot
        :type output_format: str (png, pdf, ps)
        :param show: if true, will open the new plot with the default viewer
        :type show: bool
        """
        Log.info('Plotting...')
        now = time.localtime()
        output_date = time.strftime("%Y%m%d_%H%M", now)
        output_file = os.path.join(BasicConfig.LOCAL_ROOT_DIR, expid, "plot", expid + "_" + output_date + "." +
                                   output_format)
        # print(expid)
        # print(len(joblist))
        # print("Packages")
        # print(packages)
        # print("Groups")
        # print(groups)
        # print(hide_groups)
        graph = self.create_tree_list(
            expid, joblist, packages, groups, hide_groups)
        # print(graph)
        # for u in exp.get_node_list():
        #     print(u)

        # for v in exp.get_edge_list():
        #     print(str(v.get_source()) + " to " + str(v.get_destination()))
        return False
        # print(graph.source)
        # for u in graph.get_edge_list():
        #     print(u)
        # for v in graph[u]:
        #     print(v)

        Log.debug("Saving workflow plot at '{0}'", output_file)
        if output_format == "png":
            # noinspection PyUnresolvedReferences
            graph.write_png(output_file)
        elif output_format == "pdf":
            # noinspection PyUnresolvedReferences
            graph.write_pdf(output_file)
            # result = graph.create('dot', format="plain")
            # for u in result.split("\n"):
            #     splitList = u.split(" ")
            #     if len(splitList) > 1 and splitList[0] == "node":
            #         print(splitList[0] + " " + splitList[1] + " x:" + splitList[2] + " y:" + splitList[3])

            # result = graph.create(format="plain")
            # for u in result.split("\n"):
            #     # print(u)
            #     splitList = u.split(" ")
            #     # print(splitList)
            #     if len(splitList) > 1 and splitList[0] == "node":
            #         print(splitList[0] + " " + splitList[1] + " x:" + splitList[2] + " y:" + splitList[3])

        elif output_format == "ps":
            # noinspection PyUnresolvedReferences
            graph.write_ps(output_file)
        elif output_format == "svg":
            # noinspection PyUnresolvedReferences
            graph.write_svg(output_file)
        else:
            Log.error('Format {0} not supported', output_format)
            return
        Log.result('Plot created at {0}', output_file)
        if show:
            try:
                subprocess.check_call(['xdg-open', output_file])
            except subprocess.CalledProcessError:
                Log.error('File {0} could not be opened', output_file)

        self.generate_output_txt(expid, joblist, path, "default")

    def generate_output_txt(self, expid, joblist, path, classictxt=False):
        Log.info('Writing status txt...')

        now = time.localtime()
        output_date = time.strftime("%Y%m%d_%H%M", now)
        file_path = os.path.join(
            BasicConfig.LOCAL_ROOT_DIR, expid, "status", expid + "_" + output_date + ".txt")

        if not os.path.exists(os.path.dirname(file_path)):
            os.makedirs(os.path.dirname(file_path))

        output_file = open(file_path, 'w+')
        if classictxt:
            for job in joblist:
                log_out = ""
                log_err = ""
                if job.status in [Status.FAILED, Status.COMPLETED]:
                    log_out = path + "/" + job.local_logs[0]
                    log_err = path + "/" + job.local_logs[1]

                output = job.name + " " + \
                    Status().VALUE_TO_KEY[job.status] + \
                    " " + log_out + " " + log_err + "\n"
                output_file.write(output)
        else:
            output_file.write(
                "Writing jobs, they're grouped by [FC and DATE] \n")
            self.write_output_txt_recursive(
                joblist[0], output_file, "", file_path)
            output_file.close()
        Log.result('Status txt created at {0}', output_file)

    def write_output_txt_recursive(self, job, output_file, level, path):
        log_out = ""
        log_err = ""
        # if job.status in [Status.FAILED, Status.COMPLETED]:
        #    log_out = path + "/" + job.local_logs[0]
        #    log_err = path + "/" + job.local_logs[1]
        # + " " + log_out + " " + log_err + "\n"
        output = level + job.name + " " + \
            Status().VALUE_TO_KEY[job.status] + "\n"
        output_file.write(output)
        if job.has_children() > 0:
            for child in job.children:
                self.write_output_txt_recursive(
                    child, output_file, "_" + level, path)

    def generate_output_stats(self, expid, joblist, output_format="pdf", period_ini=None, period_fi=None, show=False):
        """
        Plots stats for joblist and stores it in a file

        :param expid: experiment's identifier
        :type expid: str
        :param joblist: joblist to plot
        :type joblist: JobList
        :param output_format: file format for plot
        :type output_format: str (png, pdf, ps)
        :param period_ini: initial datetime of filtered period
        :type period_ini: datetime
        :param period_fi: final datetime of filtered period
        :type period_fi: datetime
        :param show: if true, will open the new plot with the default viewer
        :type show: bool
        """
        Log.info('Creating stats file')
        now = time.localtime()
        output_date = time.strftime("%Y%m%d_%H%M", now)

        directory = os.path.join(BasicConfig.LOCAL_ROOT_DIR, expid, "stats")
        if not os.path.exists(directory):
            os.makedirs(directory)

        output_file = os.path.join(BasicConfig.LOCAL_ROOT_DIR, expid, "stats", expid + "_statistics_" + output_date +
                                   "." + output_format)

        # create_bar_diagram(expid, joblist, self.get_general_stats(expid), output_file, period_ini, period_fi)
        Log.result('Stats created at {0}', output_file)
        if show:
            try:
                subprocess.check_call(['xdg-open', output_file])
            except subprocess.CalledProcessError:
                Log.error('File {0} could not be opened', output_file)

    @staticmethod
    def clean_plot(expid):
        """
        Function to clean space on BasicConfig.LOCAL_ROOT_DIR/plot directory.
        Removes all plots except last two.

        :param expid: experiment's identifier
        :type expid: str
        """
        search_dir = os.path.join(BasicConfig.LOCAL_ROOT_DIR, expid, "plot")
        chdir(search_dir)
        files = filter(path.isfile, listdir(search_dir))
        files = [path.join(search_dir, f)
                 for f in files if 'statistics' not in f]
        files.sort(key=lambda x: path.getmtime(x))
        remain = files[-2:]
        filelist = [f for f in files if f not in remain]
        for f in filelist:
            remove(f)
        Log.result("Plots cleaned!\nLast two plots remanining there.\n")

    @staticmethod
    def clean_stats(expid):
        """
        Function to clean space on BasicConfig.LOCAL_ROOT_DIR/plot directory.
        Removes all stats' plots except last two.

        :param expid: experiment's identifier
        :type expid: str
        """
        search_dir = os.path.join(BasicConfig.LOCAL_ROOT_DIR, expid, "plot")
        chdir(search_dir)
        files = filter(path.isfile, listdir(search_dir))
        files = [path.join(search_dir, f) for f in files if 'statistics' in f]
        files.sort(key=lambda x: path.getmtime(x))
        remain = files[-1:]
        filelist = [f for f in files if f not in remain]
        for f in filelist:
            remove(f)
        Log.result("Stats cleaned!\nLast stats' plot remanining there.\n")

    @staticmethod
    def get_general_stats(expid):
        general_stats = []
        general_stats_path = os.path.join(
            BasicConfig.LOCAL_ROOT_DIR, expid, "tmp", expid + "_GENERAL_STATS")
        parser = AutosubmitConfig.get_parser(
            ConfigParserFactory(), general_stats_path)
        for section in parser.sections():
            general_stats.append((section, ''))
            general_stats += parser.items(section)
        return general_stats



#!/usr/bin/env python

# Copyright 2015 Earth Sciences Department, BSC-CNS

# This file is part of Autosubmit.

# Autosubmit is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Autosubmit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Autosubmit.  If not, see <http://www.gnu.org/licenses/>.

import smtplib
import email.utils
from email.mime.text import MIMEText
from bscearth.utils.log import Log


class MailNotifier:
    def __init__(self, basic_config):
        self.config = basic_config

    def notify_status_change(self, exp_id, job_name, prev_status, status, mail_to):
        message_text = self._generate_message_text(exp_id, job_name, prev_status, status)
        message = MIMEText(message_text)
        message['From'] = email.utils.formataddr(('Autosubmit', self.config.MAIL_FROM))
        message['Subject'] = '[Autosubmit] The job {0} status has changed to {1}'.format(job_name, str(status))
        for mail in mail_to:
            message['To'] = email.utils.formataddr((mail, mail))
            try:
                self._send_mail(self.config.MAIL_FROM, mail, message)
            except:
                Log.warning('An error occurred while sending a mail for the job {0}', job_name)

    def _send_mail(self, mail_from, mail_to, message):
        server = smtplib.SMTP_SSL(self.config.SMTP_SERVER)
        server.sendmail(mail_from, mail_to, message.as_string())
        server.quit()

    @staticmethod
    def _generate_message_text(exp_id, job_name, prev_status, status):
        return 'Autosubmit notification\n' \
               '-------------------------\n\n' \
                'Experiment id:  ' + str(exp_id) + '\n\n' \
                + 'Job name:  ' + str(job_name) + '\n\n' \
                'The job status has changed from: ' + str(prev_status) + ' to ' + str(status) + '\n\n\n\n\n' \
                'INFO: This message was auto generated by Autosubmit, '\
                'remember that you can disable these messages on Autosubmit config file. \n'



from unittest import TestCase
from autosubmit.config.config_common import AutosubmitConfig
from bscearth.utils.config_parser import ConfigParserFactory, ConfigParser
from mock import Mock
from mock import patch
from mock import mock_open
import os
import sys
from datetime import datetime

# compatibility with both versions (2 & 3)
from sys import version_info

if version_info.major == 2:
    import __builtin__ as builtins
else:
    import builtins


class TestAutosubmitConfig(TestCase):
    any_expid = 'a000'

    # dummy values for tests
    section = 'any-section'
    option = 'any-option'

    def setUp(self):
        self.config = AutosubmitConfig(self.any_expid, FakeBasicConfig, ConfigParserFactory())
        self.config.reload()

    def test_get_parser(self):
        # arrange
        file_path = 'dummy/file/path'

        parser_mock = Mock(spec=ConfigParser)
        parser_mock.read = Mock()

        factory_mock = Mock(spec=ConfigParserFactory)
        factory_mock.create_parser = Mock(return_value=parser_mock)

        # act
        returned_parser = AutosubmitConfig.get_parser(factory_mock, file_path)

        # assert
        self.assertTrue(isinstance(returned_parser, ConfigParser))
        factory_mock.create_parser.assert_called_with()
        parser_mock.read.assert_called_with(file_path)

    def test_experiment_file(self):
        self.assertEqual(self.config.experiment_file,
                         os.path.join(FakeBasicConfig.LOCAL_ROOT_DIR, self.any_expid, "conf",
                                      "expdef_" + self.any_expid + ".conf"))

    def test_platforms_parser(self):
        self.assertTrue(isinstance(self.config.platforms_parser, ConfigParser))

    def test_platforms_file(self):
        self.assertEqual(self.config.platforms_file,
                         os.path.join(FakeBasicConfig.LOCAL_ROOT_DIR, self.any_expid, "conf",
                                      "platforms_" + self.any_expid + ".conf"))

    def test_project_file(self):
        self.assertEqual(self.config.project_file,
                         os.path.join(FakeBasicConfig.LOCAL_ROOT_DIR, self.any_expid, "conf",
                                      "proj_" + self.any_expid + ".conf"))

    def test_jobs_file(self):
        self.assertEqual(self.config.jobs_file,
                         os.path.join(FakeBasicConfig.LOCAL_ROOT_DIR, self.any_expid, "conf",
                                      "jobs_" + self.any_expid + ".conf"))

    def test_get_project_dir(self):
        parser_mock = Mock(spec=ConfigParser)
        parser_mock.get = Mock(side_effect=['/dummy/path'])

        factory_mock = Mock(spec=ConfigParserFactory)
        factory_mock.create_parser = Mock(return_value=parser_mock)

        config = AutosubmitConfig(self.any_expid, FakeBasicConfig, factory_mock)
        config.reload()

        # act
        returned_project_dir = config.get_project_dir()

        # assert
        self.assertEquals(os.path.join(FakeBasicConfig.LOCAL_ROOT_DIR, self.any_expid, FakeBasicConfig.LOCAL_PROJ_DIR,
                                       '/dummy/path'), returned_project_dir)

    def test_get_wallclock(self):
        # arrange
        expected_value = '00:05'
        default_value = ''
        config, parser_mock = self._arrange_config(expected_value)
        # act
        returned_value = config.get_wallclock(self.section)
        # assert
        self._assert_get_option(parser_mock, 'WALLCLOCK', expected_value, returned_value, default_value, str)

    def test_get_processors(self):
        # arrange
        expected_value = '99999'
        default_value = 1
        config, parser_mock = self._arrange_config(expected_value)
        # act
        returned_value = config.get_processors(self.section)
        # assert
        self._assert_get_option(parser_mock, 'PROCESSORS', expected_value, returned_value, default_value, str)

    def test_get_threads(self):
        # arrange
        expected_value = '99999'
        default_value = 1
        config, parser_mock = self._arrange_config(expected_value)
        # act
        returned_value = config.get_threads(self.section)
        # assert
        self._assert_get_option(parser_mock, 'THREADS', expected_value, returned_value, default_value, str)

    def test_get_tasks(self):
        # arrange
        expected_value = '99999'
        default_value = 0
        config, parser_mock = self._arrange_config(expected_value)
        # act
        returned_value = config.get_tasks(self.section)
        # assert
        self._assert_get_option(parser_mock, 'TASKS', expected_value, returned_value, default_value, str)

    def test_get_memory(self):
        # arrange
        expected_value = '99999'
        default_value = ''
        config, parser_mock = self._arrange_config(expected_value)
        # act
        returned_value = config.get_memory(self.section)
        # assert
        self._assert_get_option(parser_mock, 'MEMORY', expected_value, returned_value, default_value, str)

    def test_that_reload_must_load_parsers(self):
        # arrange
        config = AutosubmitConfig(self.any_expid, FakeBasicConfig, ConfigParserFactory())
        parsers = ['_conf_parser', '_platforms_parser', '_jobs_parser', '_exp_parser', '_proj_parser']

        # pre-act assertions
        for parser in parsers:
            self.assertTrue(hasattr(config, parser))
            self.assertIsNone(getattr(config, parser))

        # act
        config.reload()

        # assert
        # TODO: could be improved asserting that the methods are called
        for parser in parsers:
            self.assertTrue(hasattr(config, parser))
            self.assertTrue(isinstance(getattr(config, parser), ConfigParser))

    def test_set_expid(self):
        # arrange
        config = AutosubmitConfig(self.any_expid, FakeBasicConfig, ConfigParserFactory())

        open_mock = mock_open(read_data="EXPID = dummy")
        with patch.object(builtins, "open", open_mock):
            # act
            config.set_expid('dummy-expid')

        # assert
        open_mock.assert_any_call(config.experiment_file, 'w')
        open_mock.assert_any_call(getattr(config, '_conf_parser_file'), 'w')

    def test_set_platform(self):
        # arrange
        config = AutosubmitConfig(self.any_expid, FakeBasicConfig, ConfigParserFactory())

        open_mock = mock_open(read_data="HPCARCH = dummy")
        with patch.object(builtins, "open", open_mock):
            # act
            config.set_platform('dummy-platform')

        # assert
        open_mock.assert_any_call(config.experiment_file, 'w')

    def test_set_version(self):
        # arrange
        config = AutosubmitConfig(self.any_expid, FakeBasicConfig, ConfigParserFactory())

        open_mock = mock_open(read_data='AUTOSUBMIT_VERSION = dummy')
        with patch.object(builtins, "open", open_mock):
            # act
            config.set_version('dummy-vesion')

        # assert
        open_mock.assert_any_call(getattr(config, '_conf_parser_file'), 'w')

    def test_set_safetysleeptime(self):
        # arrange
        config = AutosubmitConfig(self.any_expid, FakeBasicConfig, ConfigParserFactory())

        open_mock = mock_open(read_data='SAFETYSLEEPTIME = dummy')
        with patch.object(builtins, "open", open_mock):
            # act
            config.set_safetysleeptime(999999)

        # assert
        open_mock.assert_any_call(getattr(config, '_conf_parser_file'), 'w')

    def test_load_project_parameters(self):
        # arrange
        parser_mock = Mock(spec=ConfigParser)
        parser_mock.sections = Mock(return_value=['DUMMY_SECTION_1', 'DUMMY_SECTION_2'])
        parser_mock.items = Mock(side_effect=[[['dummy_key1', 'dummy_value1'], ['dummy_key2', 'dummy_value2']],
                                              [['dummy_key3', 'dummy_value3'], ['dummy_key4', 'dummy_value4']]])

        factory_mock = Mock(spec=ConfigParserFactory)
        factory_mock.create_parser = Mock(return_value=parser_mock)

        config = AutosubmitConfig(self.any_expid, FakeBasicConfig, factory_mock)
        config.reload()

        # act
        project_parameters = config.load_project_parameters()

        # assert
        parser_mock.items.assert_any_call('DUMMY_SECTION_1')
        parser_mock.items.assert_any_call('DUMMY_SECTION_2')
        self.assertEquals(4, len(project_parameters))
        for i in range(1, 4):
            self.assertEquals(project_parameters.get('dummy_key' + str(i)), 'dummy_value' + str(i))

    def test_get_startdates_list(self):
        # arrange
        parser_mock = Mock(spec=ConfigParser)
        # TODO: Check if these are all accepted formats
        parser_mock.get = Mock(return_value='1920 193005 19400909 1950[01 0303]')

        factory_mock = Mock(spec=ConfigParserFactory)
        factory_mock.create_parser = Mock(return_value=parser_mock)

        config = AutosubmitConfig(self.any_expid, FakeBasicConfig, factory_mock)
        config.reload()

        # act
        returned_dates = config.get_date_list()

        # assert
        self.assertEquals(5, len(returned_dates))
        self.assertTrue(datetime(1920, 1, 1) in returned_dates)
        self.assertTrue(datetime(1930, 5, 1) in returned_dates)
        self.assertTrue(datetime(1940, 9, 9) in returned_dates)
        self.assertTrue(datetime(1950, 1, 1) in returned_dates)
        self.assertTrue(datetime(1950, 3, 3) in returned_dates)

    def test_check_project(self):
        # arrange
        parser_mock = Mock(spec=ConfigParser)
        parser_mock.read = Mock(side_effect=Exception)

        factory_mock = Mock(spec=ConfigParserFactory)
        factory_mock.create_parser = Mock(return_value=parser_mock)

        config = AutosubmitConfig(self.any_expid, FakeBasicConfig, ConfigParserFactory())
        config2 = AutosubmitConfig(self.any_expid, FakeBasicConfig, ConfigParserFactory())
        config3 = AutosubmitConfig(self.any_expid, FakeBasicConfig, factory_mock)

        config._proj_parser_file = ''

        # act
        should_be_true = config.check_proj()
        should_be_true2 = config2.check_proj()
        should_be_false = config3.check_proj()

        # assert
        self.assertTrue(should_be_true)
        self.assertEquals(None, config._proj_parser)
        self.assertTrue(should_be_true2)
        self.assertFalse(should_be_false)

    def test_load_parameters(self):
        # arrange
        parser_mock = Mock(spec=ConfigParser)
        parser_mock.sections = Mock(side_effect=[['dummy-section1'], ['dummy-section2'], ['dummy-section3']])

        parser_mock.options = Mock(side_effect=[['dummy-option1', 'dummy-option2'],
                                                ['dummy-option3', 'dummy-option4']])

        parser_mock.get = Mock(return_value='dummy-value')

        parser_mock.items = Mock(return_value=[['dummy-key1', 'dummy-value1'],
                                               ['dummy-key2', 'dummy-value2']])

        factory_mock = Mock(spec=ConfigParserFactory)
        factory_mock.create_parser = Mock(return_value=parser_mock)

        config = AutosubmitConfig(self.any_expid, FakeBasicConfig, factory_mock)
        config.reload()

        # act
        returned_parameters = config.load_parameters()

        # assert
        self.assertEquals(6, len(returned_parameters))
        self.assertTrue(returned_parameters.has_key('dummy-option1'))
        self.assertTrue(returned_parameters.has_key('dummy-option2'))
        self.assertTrue(returned_parameters.has_key('dummy-option3'))
        self.assertTrue(returned_parameters.has_key('dummy-option4'))
        self.assertTrue(returned_parameters.has_key('dummy-key1'))
        self.assertTrue(returned_parameters.has_key('dummy-key2'))

    def test_git_project_commit(self):
        # arrange
        # noinspection PyPep8Naming
        sys.modules['subprocess'].CalledProcessError = Exception
        sys.modules['subprocess'].check_output = Mock(side_effect=[Exception,
                                                                   'dummy/path/', Exception,
                                                                   'dummy/path/', 'dummy/sha/'])
        parser_mock = Mock(spec=ConfigParser)

        factory_mock = Mock(spec=ConfigParserFactory)
        factory_mock.create_parser = Mock(return_value=parser_mock)

        config = AutosubmitConfig(self.any_expid, FakeBasicConfig, factory_mock)
        config.reload()

        # TODO: reorganize act & improve the assertions
        should_be_false = config.set_git_project_commit(config)
        should_be_false2 = config.set_git_project_commit(config)

        open_mock = mock_open(read_data='PROJECT_BRANCH = dummy \n PROJECT_COMMIT = dummy')
        with patch.object(builtins, "open", open_mock):
            # act
            should_be_true = config.set_git_project_commit(config)

            # assert
            self.assertTrue(should_be_true)

        self.assertFalse(should_be_false)
        self.assertFalse(should_be_false2)

    # TODO: Test specific cases
    def test_check_jobs_conf(self):
        # arrange
        parser_mock = Mock(spec=ConfigParser)
        parser_mock.sections = Mock(side_effect=[['dummy-section1', 'dummy-section2'],
                                                 ['dummy-platform1', 'dummy-platform2']])
        parser_mock.has = Mock(return_value=True)

        parser_mock.get = Mock(side_effect=['true', 'dummy-platform1', 'dependency-1 dependency-2',
                                            'dependency-1 dependency-2', 'once',
                                            'true', 'dummy-platform1', 'dependency-1 dependency-2',
                                            'dependency-1 dependency-2', 'once'])

        factory_mock = Mock(spec=ConfigParserFactory)
        factory_mock.create_parser = Mock(return_value=parser_mock)

        config = AutosubmitConfig(self.any_expid, FakeBasicConfig, factory_mock)
        config.reload()

        # act
        should_be_true = config.check_jobs_conf()

        # assert
        self.assertTrue(should_be_true)

    # TODO: Test specific cases
    def test_check_platforms_conf(self):
        # arrange
        parser_mock = Mock(spec=ConfigParser)
        parser_mock.sections = Mock(side_effect=[[], [], ['dummy-section1'], ['dummy-section1', 'dummy-section2']])
        parser_mock.has = Mock(return_value=True)

        parser_mock.get = Mock(side_effect=['not-ps', 'true', 'false', 111, 222,
                                            'not-ps', 'true', 'false', 111, 222])

        factory_mock = Mock(spec=ConfigParserFactory)
        factory_mock.create_parser = Mock(return_value=parser_mock)

        config = AutosubmitConfig(self.any_expid, FakeBasicConfig, factory_mock)
        config.reload()

        # act
        should_be_true = config.check_platforms_conf()

        # assert
        self.assertTrue(should_be_true)

    def test_check_conf_files(self):
        # arrange
        truth_mock = Mock(return_value=True)

        config = AutosubmitConfig(self.any_expid, FakeBasicConfig, ConfigParserFactory())
        config.reload()
        config.check_autosubmit_conf = truth_mock
        config.check_platforms_conf = truth_mock
        config.check_jobs_conf = truth_mock
        config.check_expdef_conf = truth_mock

        config2 = AutosubmitConfig(self.any_expid, FakeBasicConfig, ConfigParserFactory())
        config2.reload()
        config2.check_autosubmit_conf = truth_mock
        config2.check_platforms_conf = truth_mock
        config2.check_jobs_conf = truth_mock
        config2.check_expdef_conf = Mock(return_value=False)

        # act
        should_be_true = config.check_conf_files()
        should_be_false = config2.check_conf_files()

        # assert
        self.assertTrue(should_be_true)
        self.assertFalse(should_be_false)
        self.assertEquals(7, truth_mock.call_count)

    def test_is_valid_mail_with_non_mail_address_returns_false(self):
        self.assertFalse(AutosubmitConfig.is_valid_mail_address('12345'))

    def test_is_valid_mail_with_mail_address_returns_true(self):
        self.assertTrue(AutosubmitConfig.is_valid_mail_address('example@example.org'))

    #############################
    ## Helper functions & classes

    def _assert_get_option(self, parser_mock, option, expected_value, returned_value, default_value, expected_type):
        self.assertTrue(isinstance(returned_value, expected_type))
        self.assertEqual(expected_value, returned_value)
        parser_mock.get_option.assert_called_once_with(self.section, option, default_value)

    def _arrange_config(self, option_value):
        # arrange
        parser_mock = Mock(spec=ConfigParser)
        parser_mock.has_option = Mock(return_value=True)
        parser_mock.get = Mock(return_value=option_value)
        parser_mock.get_option = Mock(return_value=option_value)
        factory_mock = Mock(spec=ConfigParserFactory)
        factory_mock.create_parser = Mock(return_value=parser_mock)
        config = AutosubmitConfig(self.any_expid, FakeBasicConfig, factory_mock)
        config.reload()
        return config, parser_mock

    def _create_parser_mock(self, has_option, returned_option=None):
        parser_mock = Mock(spec=ConfigParser)
        parser_mock.has_option = Mock(return_value=has_option)
        parser_mock.get = Mock(return_value=returned_option)
        return parser_mock


class FakeBasicConfig:
    DB_DIR = '/dummy/db/dir'
    DB_FILE = '/dummy/db/file'
    DB_PATH = '/dummy/db/path'
    LOCAL_ROOT_DIR = '/dummy/local/root/dir'
    LOCAL_TMP_DIR = '/dummy/local/temp/dir'
    LOCAL_PROJ_DIR = '/dummy/local/proj/dir'
    DEFAULT_PLATFORMS_CONF = ''
    DEFAULT_JOBS_CONF = ''




import xml.etree.ElementTree as ET
import traceback

def get_cas_user_from_xml(xmlstring):    
    """
    Parses xmlstring and looks for the user tag.
    
    :return: user  
    :rtype: str or None
    """
    try:
      root_node = ET.fromstring(xmlstring)
      user = None
      for child in root_node:        
          if child.tag == '{http://www.yale.edu/tp/cas}authenticationSuccess':
              for subchild in child:
                  if subchild.tag == '{http://www.yale.edu/tp/cas}user':
                      user = subchild.text
    except Exception as exp:
      print(exp)
      print(traceback.format_exc)
    return user


#!/usr/bin/env python

# Copyright 2017 Earth Sciences Department, BSC-CNS

# This file is part of Autosubmit.

# Autosubmit is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Autosubmit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Autosubmit.  If not, see <http://www.gnu.org/licenses/>.

import os

from autosubmitAPIwu.platforms.paramiko_platform import ParamikoPlatform, ParamikoPlatformException
from bscearth.utils.log import Log

from autosubmitAPIwu.platforms.headers.pbs10_header import Pbs10Header
from autosubmitAPIwu.platforms.headers.pbs11_header import Pbs11Header
from autosubmitAPIwu.platforms.headers.pbs12_header import Pbs12Header


class PBSPlatform(ParamikoPlatform):
    """
    Class to manage jobs to host using PBS scheduler

    :param expid: experiment's identifier
    :type expid: str
    :param version: scheduler version
    :type version: str
    """

    def __init__(self, expid, name, config, version):
        ParamikoPlatform.__init__(self, expid, name, config)
        self._version = version

        if str.startswith(version, '10'):
            self._header = Pbs10Header()
        elif str.startswith(version, '11'):
            self._header = Pbs11Header()
        elif str.startswith(version, '12'):
            self._header = Pbs12Header()
        else:
            Log.error('PBS version {0} not supported'.format(version))
            raise HPCPlatformException(
                'PBS version {0} not supported'.format(version))

        self.job_status = dict()
        self.job_status['COMPLETED'] = ['F', 'E', 'c', 'C']
        self.job_status['RUNNING'] = ['R']
        self.job_status['QUEUING'] = ['Q', 'H', 'S', 'T', 'W', 'U', 'M']
        self.job_status['FAILED'] = ['Failed', 'Node_fail', 'Timeout']
        # No ned to write from API
        # self.update_cmds()

    def update_cmds(self):
        """
        Updates commands for platforms
        """
        self.root_dir = os.path.join(
            self.scratch, self.project, self.user, self.expid)
        self.remote_log_dir = os.path.join(self.root_dir, "LOG_" + self.expid)
        self.cancel_cmd = "ssh " + self.host + " qdel"
        self._checkhost_cmd = "ssh " + self.host + " echo 1"
        self.put_cmd = "scp"
        self.get_cmd = "scp"
        self.mkdir_cmd = "ssh " + self.host + " mkdir -p " + self.remote_log_dir
        self._submit_cmd = "ssh " + self.host + " qsub -d " + \
            self.remote_log_dir + " " + self.remote_log_dir + "/ "

        if str.startswith(self._version, '11'):
            self._checkjob_cmd = "ssh " + self.host + " qstat"

    def get_checkhost_cmd(self):
        return self._checkhost_cmd

    def get_remote_log_dir(self):
        return self.remote_log_dir

    def get_mkdir_cmd(self):
        return self.mkdir_cmd

    def parse_job_output(self, output):
        # job_state = output.split('\n')[2].split()[4]
        # return job_state
        return output

    def get_submitted_job_id(self, output):
        return output.split('.')[0]

    def jobs_in_queue(self):
        return ''.split()

    def get_submit_cmd(self, job_script, job):
        return self._submit_cmd + job_script

    def get_checkjob_cmd(self, job_id):
        if str.startswith(self._version, '11'):
            return self._checkjob_cmd + str(job_id)
        else:
            return "ssh " + self.host + " " + self.get_qstatjob(job_id)



#!/usr/bin/env python

# Copyright 2017 Earth Sciences Department, BSC-CNS

# This file is part of Autosubmit.

# Autosubmit is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Autosubmit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Autosubmit.  If not, see <http://www.gnu.org/licenses/>.

import networkx
import datetime
import time

from networkx.algorithms.dag import is_directed_acyclic_graph
from networkx import DiGraph
from networkx import dfs_edges
from networkx import NetworkXError


def transitive_reduction(graph):
    try:
        return networkx.algorithms.dag.transitive_reduction(graph)
    except Exception as exp:
        return None
    # if not is_directed_acyclic_graph(graph):
    #     raise NetworkXError("Transitive reduction only uniquely defined on directed acyclic graphs.")
    # reduced_graph = DiGraph()
    # reduced_graph.add_nodes_from(graph.nodes())
    # for u in graph:
    #     u_edges = set(graph[u])
    #     for v in graph[u]:
    #         u_edges -= {y for x, y in dfs_edges(graph, v)}
    #     reduced_graph.add_edges_from((u, v) for v in u_edges)
    # return reduced_graph


class Dependency(object):
    """
    Class to manage the metadata related with a dependency

    """

    def __init__(self, section, distance=None, running=None, sign=None, delay=-1, splits=None, select_chunks=list()):
        self.section = section
        self.distance = distance
        self.running = running
        self.sign = sign
        self.delay = delay
        self.splits = splits
        self.select_chunks_dest = list()
        self.select_chunks_orig = list()
        for chunk_relation in select_chunks:
            self.select_chunks_dest.append(chunk_relation[0])
            if len(chunk_relation) > 1:
                self.select_chunks_orig.append(chunk_relation[1])
            else:
                self.select_chunks_orig.append([])


class SimpleJob(object):
    """
    A simple replacement for jobs
    """

    def __init__(self, name, tmppath, statuscode):
        self.name = name
        self._tmp_path = tmppath
        self.status = statuscode


class SubJob(object):
    """
    Class to manage package times
    """

    def __init__(self, name, package=None, queue=0, run=0, total=0, status="UNKNOWN"):
        self.name = name
        self.package = package
        self.queue = queue
        self.run = run
        self.total = total
        self.status = status
        self.transit = 0
        self.parents = list()
        self.children = list()


class SubJobManager(object):
    """
    Class to manage list of SubJobs
    """

    def __init__(self, subjoblist, job_to_package=None, package_to_jobs=None, current_structure=None):
        self.subjobList = subjoblist
        self.job_to_package = job_to_package
        self.package_to_jobs = package_to_jobs
        self.current_structure = current_structure
        self.subjobindex = dict()
        self.subjobfixes = dict()
        self.process_index()
        self.process_times()

    def process_index(self):
        """
        Builds a dictionary of jobname -> SubJob object. 
        """
        for subjob in self.subjobList:
            self.subjobindex[subjob.name] = subjob

    def process_times(self):
        """
        """
        if (self.job_to_package) and (self.package_to_jobs):
            if(self.current_structure) and len(self.current_structure.keys()) > 0:
                # Structure exists
                new_queues = dict()
                fixes_applied = dict()
                for package in self.package_to_jobs:
                    # SubJobs in Package
                    local_structure = dict()
                    # SubJob Name -> SubJob Object
                    local_index = dict()
                    subjobs_in_package = filter(lambda x: x.package ==
                                                package, self.subjobList)
                    local_jobs_in_package = [job for job in subjobs_in_package]
                    # Build index
                    for sub in local_jobs_in_package:
                        local_index[sub.name] = sub
                    # Build structure
                    for sub_job in local_jobs_in_package:
                        # If job in current_structure, store children names in dictionary
                        # local_structure: Job Name -> Children (if present in the Job package)
                        local_structure[sub_job.name] = [v for v in self.current_structure[sub_job.name]
                                                         if v in self.package_to_jobs[package]] if sub_job.name in self.current_structure else list()
                        # Assign children to SubJob in local_jobs_in_package
                        sub_job.children = local_structure[sub_job.name]
                        # Assign sub_job Name as a parent of each of its children
                        for child in local_structure[sub_job.name]:
                            local_index[child].parents.append(sub_job.name)

                    # Identify root as the job with no parents in the package
                    roots = [sub for sub in local_jobs_in_package if len(
                        sub.parents) == 0]

                    # While roots exists (consider pop)
                    while(len(roots) > 0):
                        sub = roots.pop(0)
                        if len(sub.children) > 0:
                            for sub_children_name in sub.children:
                                if sub_children_name not in new_queues:
                                    # Add children to root to continue the sequence of fixes
                                    roots.append(
                                        local_index[sub_children_name])
                                    fix_size = self.subjobindex[sub.name].queue + \
                                        self.subjobindex[sub.name].run
                                    # fixes_applied.setdefault(sub_children_name, []).append(fix_size) # If we care about repetition
                                    # Retain the greater fix size
                                    if fix_size > fixes_applied.get(sub_children_name, 0):
                                        fixes_applied[sub_children_name] = fix_size
                                    fixed_queue_time = self.subjobindex[sub_children_name].queue - fix_size
                                    new_queues[sub_children_name] = fixed_queue_time if fixed_queue_time >= 0 else 0
                                    # print(new_queues[sub_name])

                for key, value in new_queues.items():
                    self.subjobindex[key].queue = value
                for name in fixes_applied:
                    self.subjobfixes[name] = fixes_applied[name]

            else:
                # There is no structure
                for package in self.package_to_jobs:
                    # Filter only jobs in the current package
                    filtered = filter(lambda x: x.package ==
                                      package, self.subjobList)
                    # Order jobs by total time (queue + run)
                    filtered = sorted(
                        filtered, key=lambda x: x.total, reverse=False)
                    # Sizes of fixes
                    fixes_applied = dict()
                    if len(filtered) > 1:
                        temp_index = 0
                        filtered[0].transit = 0
                        # Reverse for
                        for i in range(len(filtered) - 1, 0, -1):
                            # Assume that the total time of the next job is always smaller than
                            # the queue time of the current job
                            # because the queue time of the current also considers the
                            # total time of the previous (next because of reversed for) job by default
                            # Confusing? It is.
                            # Assign to transit the adjusted queue time
                            filtered[i].transit = filtered[i].queue - \
                                filtered[i - 1].total

                        # Positive or zero transit time
                        positive = len(
                            [job for job in filtered if job.transit >= 0])

                        if (positive > 1):
                            for i in range(0, len(filtered)):
                                if filtered[i].transit >= 0:
                                    temp_index = i
                                    if i > 0:
                                        # Only consider after the first job
                                        filtered[i].queue = (filtered[i].queue -
                                                             filtered[i - 1].total) if (filtered[i].queue -
                                                                                        filtered[i - 1].total) >= 0 else 0
                                        fixes_applied[filtered[i].name] = filtered[i - 1].total
                                else:
                                    filtered[i].queue = (filtered[i].queue -
                                                         filtered[temp_index].total) if (filtered[i].queue -
                                                                                         filtered[temp_index].total) >= 0 else 0
                                    fixes_applied[filtered[i].name] = filtered[temp_index].total
                                # it is starting of level

                    for sub in filtered:
                        self.subjobindex[sub.name].queue = sub.queue
                    for name in fixes_applied:
                        self.subjobfixes[name] = fixes_applied[name]

    def get_subjoblist(self):
        """
        Returns the list of SubJob objects with their corrected queue times
        in the case of jobs that belong to a wrapper.
        """
        return self.subjobList

    def get_collection_of_fixes_applied(self):
        """

        """
        return self.subjobfixes


def parse_output_number(self, string_number):
    """
    Parses number in format 1.0K 1.0M 1.0G

    :param string_number: String representation of number
    :type string_number: str
    :return: number in float format
    :rtype: float
    """
    number = 0.0
    if (string_number):
        last_letter = string_number.strip()[-1]
        multiplier = 1
        if last_letter == "G":
            multiplier = 1000000000
            number = string_number[:-1]
        elif last_letter == "M":
            multiplier = 1000000
            number = string_number[:-1]
        elif last_letter == "K":
            multiplier = 1000
            number = string_number[:-1]
        else:
            number = string_number
        try:
            number = float(number) * multiplier
        except Exception as exp:
            number = 0.0
            pass
    return number


def job_times_to_text(minutes_queue, minutes_running, status):
    """
    Return text correpsonding to queue and running time
    :param minutes_queue: seconds queuing (actually using seconds)  
    :type minutes_queue: int
    :param minutes_running: seconds running (actually using seconds)  
    :type minutes_running: int
    :param status: current status
    :type status: string
    :return: string
    """
    if status in ["COMPLETED", "FAILED", "RUNNING"]:
        running_text = "( " + str(datetime.timedelta(seconds=minutes_queue)) + \
            " ) + " + \
            str(datetime.timedelta(seconds=minutes_running))
    elif status in ["SUBMITTED", "QUEUING", "HELD", "HOLD"]:
        running_text = "( " + \
            str(datetime.timedelta(seconds=minutes_queue)) + " )"
    elif status in ["NA"]:
        running_text = " <small><i><b>NA</b></i></small>"
    else:
        running_text = ""

    if status == "SUSPICIOUS":
        running_text = running_text + \
            " <small><i><b>SUSPICIOUS</b></i></small>"
    return running_text


def datechunk_to_year(chunk_unit, chunk_size):
    """
    Gets chunk unit and size and returns the value in years

    :return: years  
    :rtype: double
    """
    chunk_size = chunk_size * 1.0
    options = ["year", "month", "day", "hour"]
    if (chunk_unit == "year"):
        return chunk_size
    elif (chunk_unit == "month"):
        return chunk_size / 12
    elif (chunk_unit == "day"):
        return chunk_size / 365
    elif (chunk_unit == "hour"):
        return chunk_size / 8760
    else:
        return 0


def tostamp(string_date):
    """
    String datetime to timestamp
    """
    if string_date and len(string_date) > 0:
        return int(time.mktime(datetime.datetime.strptime(string_date,
                                                          "%Y-%m-%d %H:%M:%S").timetuple()))
    else:
        return 0


def calculate_SYPD_perjob(chunk_unit, chunk_size, job_chunk, run_time):
    """
    :param chunk_unit: 
    :param chunk_size: 
    :param job_chunk: 
    :param run_time:
    """
    if job_chunk and job_chunk > 0:
        years_per_sim = datechunk_to_year(chunk_unit, chunk_size)
        return round(years_per_sim * 86400 / run_time, 2) if run_time and run_time > 0 else 0
    return None


def calculate_ASYPD_perjob(chunk_unit, chunk_size, job_chunk, queue_run_time, average_post):
    """
    :param chunk_unit: 
    :param chunk_size: 
    :param job_chunk: 
    :param queue_run_time: 
    :param average_post:
    """
    if job_chunk and job_chunk > 0:
        years_per_sim = datechunk_to_year(chunk_unit, chunk_size)
        return round(years_per_sim * 86400.0 / (queue_run_time + average_post), 2) if queue_run_time and average_post and queue_run_time > 0 and average_post > 0 else 0
    return None


def getTitle(job_name, status_color, status_text):
    return job_name + " <span class='badge' style='background-color: " + status_color + "'>#" + status_text + "</span>"



#!/usr/bin/env python

# Copyright 2017 Earth Sciences Department, BSC-CNS

# This file is part of Autosubmit.

# Autosubmit is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Autosubmit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Autosubmit.  If not, see <http://www.gnu.org/licenses/>.

from autosubmitAPIwu.job.job_common import Status
from bscearth.utils.date import date2str
import copy

class JobGrouping(object):

    def __init__(self, group_by, jobs, job_list, expand_list=list(), expanded_status=list()):
        """
        CONSTRUCTOR:\n
        group_by: automatic, date, member, chunk.\n
        jobs: A copy of the jobs.\n
        job_list: JobList instance.\n
        expand_list: A formula that specifies the jobs that should not be grouped (or that should be expanded) e.g. "[ 19601101 [ fc0 [1 2 3 4] fc1 [1] ] 19651101 [ fc0 [16-30] ] ]".\n
        espanded_status: List of status that should not be grouped (e.g. "RUNNING, FAILED") (or that should be expanded). Comma separated or single.\n
        """        
        self.group_by = group_by
        self.jobs = jobs
        self.job_list = job_list
        self.date_format = job_list.get_date_format()
        self.expand_list = expand_list
        self.expand_status = expanded_status
        self.automatic = False
        # Key: Group, Value: Starts as a List(of String) but at the end of the process turns into String.
        self.group_status_dict = dict() 
        # List of jobs that should not be grouped as a result of parsing expand_list    
        self.ungrouped_jobs = list()

    def group_jobs(self):
        """
        Main method of the class.\n
        Groups self.jobs according to parameters set in constructor.\n
        Choose of return value is questionable.\n

        :return: Dictionary that has 2 keys.\n
        Key 1: [Jobs] Value 1: Dictionary of Key: Job, Value: Group.\n
        Key 2: [Status] Value 1: Dictionary of Key: Group, Value: Final Status
        """
        if self.expand_list:
            self._set_expanded_jobs()

        jobs_group_dict = dict()
        blacklist = list()

        groups_map = dict()
        if self.group_by == 'automatic':
            # This assignment is almost irrelevant
            self.automatic = True
            # Grouping by automatic, groups_map is Empty
            # Filling dictionary Key: Job Name, Value: Group Name
            # Perhaps the original developer forgot to include self.ungrouped_jobs in the automatic grouping
            jobs_group_dict = self._automatic_grouping(groups_map)
        else:
            self._create_groups(jobs_group_dict, self.ungrouped_jobs)

            # Retrieve dictionary of group -> status list (set) in that group
            # This dictionary was filled in self._create_groups()
            for group, statuses in self.group_status_dict.items():
                # Function that returns the status of the group
                # Provides a hierarchy, returns int 
                status = self._set_group_status(statuses)
                self.group_status_dict[group] = status

        final_jobs_group = dict()
        # Iterating the items in the Dictionary that
        # associates job -> group, dictionary filled in self._create_groups()
        for job, groups in jobs_group_dict.items():
            for group in groups:
                # This blacklist is always empty
                if group not in blacklist:
                    # This part is confusing
                    # groups_map filled if group_by is automatic, otherwise empty
                    # Iterating through old small groups
                    while group in groups_map:
                        # group = new bigger group
                        group = groups_map[group]                    
                    # Testing if group exists
                    # Status is now unique from the process above
                    if group in self.group_status_dict:
                        # Start filling final_jobs_group
                        if job not in final_jobs_group:
                            final_jobs_group[job] = list()
                        # Associate job -> group
                        final_jobs_group[job].append(group)
        # Why?
        jobs_group_dict = final_jobs_group

        groups_dict = dict()
        groups_dict['jobs'] = jobs_group_dict
        groups_dict['status'] = self.group_status_dict

        return groups_dict

    # Fill self.ungrouped_jobs 
    def _set_expanded_jobs(self):
        """
        Uses self.expand_list to produce a list of jobs that should not be grouped (or that must be expanded).\n
        Parses the input self.expand_list as a nested expression.\n
        Uses self.group_by as a decision variable.\n

        Modifies: self.expand_list
        """

        text = self.expand_list

        self.ungrouped_jobs = []

        from pyparsing import nestedExpr
        """
        Function to parse rerun specification from json format

        :param text: text to parse
        :type text: list
        :return: parsed output
        """
        count = 0

        out = nestedExpr('[', ']').parseString(text).asList()

        depth = lambda L: isinstance(L, list) and max(map(depth, L)) + 1

        if self.group_by == 'date':
            if depth(out) == 2:
                dates = list()
                for date in out[0]:
                    dates.append(date)
                self.ungrouped_jobs = dates
            else:
                raise ValueError("Please check the syntax of the expand parameter including only dates")
        elif self.group_by == 'member':
            if depth(out) == 3:
                for element in out[0]:
                    if count % 2 == 0:
                        date = out[0][count]
                        members = out[0][count + 1]
                        for member in members:
                            self.ungrouped_jobs.append(date + '_' + member)
                        count += 1
                    else:
                        count += 1
            else:
                raise ValueError(
                    "Please check the syntax of the expand parameter including dates and the corresponding members")
        elif self.group_by == 'chunk':
            if depth(out) == 4:
                for element in out[0]:
                    if count % 2 == 0:
                        date = out[0][count]
                        member_chunks = out[0][count + 1]
                        member_count = 0
                        for element_member in member_chunks:
                            if member_count % 2 == 0:
                                member = member_chunks[member_count]
                                chunks = list()
                                for chunk in member_chunks[member_count + 1]:
                                    if chunk.find("-") != -1:
                                        numbers = chunk.split("-")
                                        for count in range(int(numbers[0]), int(numbers[1]) + 1):
                                            chunks.append(count)
                                    else:
                                        chunks.append(int(chunk))
                                for chunk in chunks:
                                    self.ungrouped_jobs.append(date + '_' + member + '_' + str(chunk))
                            member_count += 1
                    count += 1
            else:
                raise ValueError(
                    "Please check the syntax of the expand parameter including dates and the corresponding members and chunks")

    def _set_group_status(self, statuses):
        """
        Receives a collection of status.\n 
        :return: Final status.\n
        :rtype: int
        """
        if isinstance(statuses, int):
            return statuses
        if len(statuses) == 1:
            return next(iter(statuses))
        else:
            if Status.FAILED in statuses:
                return Status.FAILED
            elif Status.RUNNING in statuses:
                return Status.RUNNING
            elif Status.SUBMITTED in statuses:
                return Status.SUBMITTED
            elif Status.QUEUING in statuses:
                return Status.QUEUING
            elif Status.READY in statuses:
                return Status.READY
            elif Status.WAITING in statuses:
                return Status.WAITING
            elif Status.SUSPENDED in statuses:
                return Status.SUSPENDED
            elif Status.UNKNOWN in statuses:
                return Status.UNKNOWN

    def _create_groups(self, jobs_group_dict, blacklist=list()):
        """
        Creates some an approximation of the groups.
        Fills self.group_status_dict[] as dictionary Key: Group, Value: Set of Status.\n

        jobs_group_dict: as a reference and fills it. Key: Job.Name, Value: Group.\n
        blacklist: List of jobs that should not be grouped.\n
        When called form automatic, receives en empty list as blacklist.\n
        When called from other grouping option, the blacklist starts as the list self.ungrouped_jobs.\n

        Uses self.expand_status (list of status that should not be grouped).\n
        Modifies: self.jobs, self.group_status_dict, [parameter] jobs_group_dict
        """
        # Reverse iteration for some reason, bottom-up approach perhaps
        for i in reversed(range(len(self.jobs))):
            job = self.jobs[i]

            groups = []
            # Returns True or False
            if not self._check_synchronized_job(job, groups):
                # Split is only set when doing automatic
                if self.group_by == 'split':
                    if job.split is not None:
                        idx = job.name.rfind("_")
                        groups.append(job.name[:idx - 1] + job.name[idx + 1:])                
                elif self.group_by == 'chunk':
                    if job.chunk is not None:
                        # Building group name
                        groups.append(date2str(job.date, self.date_format) + '_' + job.member + '_' + str(job.chunk))
                elif self.group_by == 'member':
                    if job.member is not None:
                        groups.append(date2str(job.date, self.date_format) + '_' + job.member)
                elif self.group_by == 'date':
                    if job.date is not None:                        
                        groups.append(date2str(job.date, self.date_format))
            # If a group has been created, then current job is used, so it is taken out of the original list.
            # Modifying an object while iterating it is not really recommended in my opinion
            if groups:
                self.jobs.pop(i)
                #print("Popping: " + str(job1.name) + " because " + str(groups))

            # Only one group is generated every iteration
            while groups:
                group = groups.pop(0)
                # Checking blacklist to avoid repetition
                if group not in blacklist:
                    if group not in self.group_status_dict:
                        self.group_status_dict[group] = set()
                    # Dictionary of group name to a list of status codes (0 -> WAITING) of its jobs
                    self.group_status_dict[group].add(job.status)                                    
                    # If status code of job in expand_status list (from input command), then it should not be grouped 
                    # OR 
                    # If automatic grouping and group already in the dictionary from above (which is redundant)
                    # AND the length of the value for the key 'group' is greater than 1, meaning that more than 1 status has been added, not allowed for automatic
                    if job.status in self.expand_status or \
                            self.automatic and group in self.group_status_dict and (len(self.group_status_dict[group]) > 1):
                        # Remove the group from the result
                        self.group_status_dict.pop(group)
                        # Adding the group to the blacklist
                        blacklist.append(group)
                        break
                    # Is this job in the list of jobs grouped?    
                    if job.name not in jobs_group_dict:
                        jobs_group_dict[job.name] = list()
                    # Dictionary of jobs to the groups they belong
                    jobs_group_dict[job.name].append(group)
                    


    # This is always false
    def _check_synchronized_job(self, job, groups):
        """
        Always returns False
        Does not change groups
        """
        synchronized = False
        # Making sure job is a chunk
        if job.chunk is not None:   
            # job.chunk exists but .date and .member don't. Is that possible?         
            if job.date is None and job.member is None:
                # Rule: If job.chunk exists, and job.date and job.member are None
                # then this job is sync
                synchronized = True
                for date in self.job_list.get_date_list():
                    # Create group name for every date in experiment
                    group_name = date2str(date, self.date_format)
                    if self.group_by in ['member', 'chunk']:                        
                        for member in self.job_list.get_member_list():
                            # If group_by is member, add +member to group name
                            group_name += '_' + member
                            if self.group_by in ['chunk']:
                                # If group_by is chunk, add +chunk to group name
                                group_name += '_' + str(job.chunk)
                            # Adding group name to list of groups
                            groups.append(group_name)
                            # Innecesary line
                            group_name = date2str(date, self.date_format)
                    else:
                        groups.append(group_name)
            # .date is not None and .member is not None
            elif job.member is None:
                # Rule: If job.chunk exists, and job.member is None, meaning that job.date is not None
                # then this job is sync
                synchronized = True                
                if self.group_by == 'date':
                    # Adding groups to list
                    groups.append(date2str(job.date, self.date_format))
                else:
                    for member in self.job_list.get_member_list():
                        group_name = date2str(job.date, self.date_format) + '_' + member
                        if self.group_by in ['chunk']:
                            group_name += '_' + str(job.chunk)
                        groups.append(group_name)

        return synchronized


    def _automatic_grouping(self, groups_map):
        """
        Performs automatic grouping.\n
        groups_map: is empty. In this process this will be filled as a Dictionary Key: Old small group, Value: New bigger group. \n
        Modifies: *self.groups_status_dict*, self.jobs\n

        :return: Dictionary that matches job name to corresponding group name.\n
        :rtype: Dictionary Key: String, Value: String
        """

        all_jobs = copy.deepcopy(self.jobs)
        # Try running as split
        split_groups, split_groups_status = self._create_splits_groups()
        
        blacklist = list()
        jobs_group_dict = dict()
        self.group_status_dict = dict()
        # Try running as chunk
        self.group_by = 'chunk'
        self.jobs = all_jobs
        self._create_groups(jobs_group_dict, blacklist)
        
        for group, statuses in self.group_status_dict.items():
            # Concludes a status for the group
            status = self._set_group_status(statuses)
            # Assign that status to the key group in the dictionary
            self.group_status_dict[group] = status


        # At this point groups_map is empty
        
        # Enter high level thing
        self._create_higher_level_group(self.group_status_dict.keys(), groups_map)

        # Using split_groups, split_groups_status that were extracted at the start of the function
        # Also uses jobs_group_dict that is a result from chunk grouping
        self._fix_splits_automatic_grouping(split_groups, split_groups_status, jobs_group_dict)


        # Check if remaining jobs can be grouped, reversed so it performs bottom-up
        # Since the last grouping was done by chunk, the list self.jobs still contains those jobs that can be merged into bigger groups
        # See self._cheate_higher_level_group
        for i in reversed(range(len(self.jobs))):            
            job = self.jobs[i]            
            for group, status in self.group_status_dict.items():
                # If the name of the group is contained in the name of the job and they have the same status
                if group in job.name and status == job.status:
                    # Add the job and assign it to the new group
                    jobs_group_dict[job.name] = [group]
                    self.jobs.pop(i)

        return jobs_group_dict


    def _create_splits_groups(self):
        """
        Only called from automatic grouping.\n
        Sets self.group_by to 'split', starts a bottom-up grouping process.\n

        :return: Dictionary of Job Names to Group Names, Dictionary of Group Names to List of Status of Jobs included in Group.\n
        :rtype: Tuple([Dictionary Key: String, Value: String], [Dictonary Key: String, Value: List(of String)])
        """
        jobs_group_dict = dict()

        self.group_by = 'split'
        self._create_groups(jobs_group_dict, list())
        return jobs_group_dict, self.group_status_dict

    def _fix_splits_automatic_grouping(self, split_groups, split_groups_status, jobs_group_dict):
        """
        This function will only be executed when jobs are using split in the experiment, and the group_by options has been set to automatic.
        """
        if split_groups and split_groups_status:
            group_maps = dict()
            for group in self.group_status_dict.keys():
                matching_groups = [split_group for split_group in split_groups_status.keys() if group in split_group]
                for matching_group in matching_groups:
                    group_maps[matching_group] = group
                    split_groups_status.pop(matching_group)

            for split_group, statuses in split_groups_status.items():
                status = self._set_group_status(statuses)
                self.group_status_dict[split_group] = status

            for job, groups in split_groups.items():
                final_groups = list()
                for group in groups:
                    if group in group_maps:
                        group = group_maps[group]
                    final_groups.append(group)
                if final_groups:
                    jobs_group_dict[job] = final_groups

    def _check_valid_group(self, groups_list, name, groups_map):
        """
        Determines if a group_list can be merged into a new bigger group name.\n
        groups_list: List of groups names of those that will possibly be grouped inside a higher group. \n
        name: Also known as new_group, is a possible new group name that groups groups_list.\n
        Modifies self.group_status_dict. Adds new bigger groups.\n
        groups_map: Starts as Empty.\n
        :return: True if the assignment is valid, False otherwise.\n
        :rtype: Boolean
        """
        # Retrieves status from dictionary Keys: group name, Values: Status
        # At this point it is a unique value per key
        group_status = self.group_status_dict[groups_list[0]]

        # Iterating starting at second item
        for group in groups_list[1:]:
            status = self.group_status_dict[group]
            # If the status of the first item in the list of existing groups 
            # is not the same as the other existing groups, then not valid            
            if status != group_status:
                return False
        
        # At this point the assignment is valid   
        for group in groups_list:
            # Take out of dict the groups that will be merged
            self.group_status_dict.pop(group)
            # map group to name (proposed group)
            groups_map[group] = name

        # Add name to group dict with the status previously determined
        self.group_status_dict[name] = group_status
        return True

    def _create_higher_level_group(self, groups_to_check, groups_map):
        """
        Reviews groups_to_check and attempts to merge these groups into bigger ones. The list of groups that can be merged is stored in groups_map.
        Curiously, it does not receive jobs_group_dict that is highly related to self.group_status_dict.\n
        Modifies self.group_status_dict.\n
        groups_to_check: Groups created from chunk pre process, only group names as it receives only keys.\n
        groups_map: Empty. Dictionary Key: Old small group, Value: New bigger group. \n
        Also uses: self.group_status_dict, self.job_list
        """
        
        checked_groups = list()
        for group in groups_to_check:
            # This if is not needed, groups_to_check === self.group_status_dict.keys()
            if group in self.group_status_dict:
                split_count = len(group.split('_'))
                # splits > 1 indicates that we are not at date level
                # split === 1 indicates date level, if date level then no more reduction is possible
                if split_count > 1:
                    # new group name is equal to the name of the group until the last occurrence of "_"
                    # This is a possible new group
                    new_group = group[:(group.rfind("_"))]
                    # If splits === 3 then we are at chunk level
                    # Else it is member level
                    # The number of possible groups according to the level (member or chunk)
                    num_groups = len(self.job_list.get_chunk_list()) if split_count == 3 else len(self.job_list.get_member_list())
                    # Controlling repetition
                    if new_group not in checked_groups:
                        checked_groups.append(new_group)
                        # This one takes the prize for weirdest one yet
                        # Selects from self.group_status_dict (group names) if key contains new_group+'_'
                        # Meaning that we are counting the number of elements in  the list of groups already defined (self.group_status_dict) 
                        # that a are a subset of new_group+'_'
                        possible_groups = [existing_group for existing_group in list(self.group_status_dict.keys()) if
                                              new_group+'_' in existing_group]

                        # If length of list from the line before is equal to the number of chunks or members
                        # Meaning that there is parity between the possible new_group and its subsets, to the number of chunks or members.
                        # chunks in the case of making possible member groups, members in the case of making possible date groups
                        if len(possible_groups) == num_groups:
                            # Remember that groups_map starts as empty
                            if self._check_valid_group(possible_groups, new_group, groups_map):
                                # Remember that groups_to_check === self.group_status_dict.keys()
                                # So, adding a new key to self.group_status_dict
                                # An attempt to make recursive, bottom-up                                
                                groups_to_check.append(new_group)


import autosubmitAPIwu.experiment.common_requests as ExperimentUtils


def main():
    """
    Process and positions graph drawing.
    """
    ExperimentUtils.process_active_graphs()


if __name__ == "__main__":
    main()
